# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from qm.pb import compiler_pb2 as qm_dot_pb_dot_compiler__pb2
from qm.pb import frontend_pb2 as qm_dot_pb_dot_frontend__pb2
from qm.pb import inc_qm_api_pb2 as qm_dot_pb_dot_inc__qm__api__pb2
from qm.pb import inc_qua_config_pb2 as qm_dot_pb_dot_inc__qua__config__pb2
from qm.pb import job_results_pb2 as qm_dot_pb_dot_job__results__pb2
from qm.pb import qm_manager_pb2 as qm_dot_pb_dot_qm__manager__pb2


class FrontendStub(object):
    """The greeting service definition.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetVersion = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/GetVersion',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
                )
        self.HealthCheck = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/HealthCheck',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.HealthCheckResponse.FromString,
                )
        self.Execute = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/Execute',
                request_serializer=qm_dot_pb_dot_frontend__pb2.ExecutionRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.ExecutionResponse.FromString,
                )
        self.AddToQueue = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/AddToQueue',
                request_serializer=qm_dot_pb_dot_frontend__pb2.AddToQueueRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.AddToQueueResponse.FromString,
                )
        self.Compile = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/Compile',
                request_serializer=qm_dot_pb_dot_frontend__pb2.CompileRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.CompileResponse.FromString,
                )
        self.AddCompiledToQueue = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/AddCompiledToQueue',
                request_serializer=qm_dot_pb_dot_frontend__pb2.AddCompiledToQueueRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.AddCompiledToQueueResponse.FromString,
                )
        self.RemovePendingJobs = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/RemovePendingJobs',
                request_serializer=qm_dot_pb_dot_frontend__pb2.JobQueryParams.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.RemovePendingJobsResponse.FromString,
                )
        self.GetPendingJobs = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/GetPendingJobs',
                request_serializer=qm_dot_pb_dot_frontend__pb2.JobQueryParams.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.GetPendingJobsResponse.FromString,
                )
        self.GetJobExecutionStatus = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/GetJobExecutionStatus',
                request_serializer=qm_dot_pb_dot_frontend__pb2.GetJobExecutionStatusRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.GetJobExecutionStatusResponse.FromString,
                )
        self.PausedStatus = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/PausedStatus',
                request_serializer=qm_dot_pb_dot_frontend__pb2.PausedStatusRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.PausedStatusResponse.FromString,
                )
        self.Resume = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/Resume',
                request_serializer=qm_dot_pb_dot_frontend__pb2.ResumeRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.ResumeResponse.FromString,
                )
        self.Halt = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/Halt',
                request_serializer=qm_dot_pb_dot_frontend__pb2.HaltRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.HaltResponse.FromString,
                )
        self.ResetDataProcessing = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/ResetDataProcessing',
                request_serializer=qm_dot_pb_dot_frontend__pb2.ResetDataProcessingRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.ResetDataProcessingResponse.FromString,
                )
        self.ValidateConfig = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/ValidateConfig',
                request_serializer=qm_dot_pb_dot_inc__qua__config__pb2.QuaConfig.SerializeToString,
                response_deserializer=qm_dot_pb_dot_compiler__pb2.ValidationResponse.FromString,
                )
        self.Init = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/Init',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.PerformQmRequest = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/PerformQmRequest',
                request_serializer=qm_dot_pb_dot_inc__qm__api__pb2.HighQmApiRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_inc__qm__api__pb2.HighQmApiResponse.FromString,
                )
        self.RequestData = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/RequestData',
                request_serializer=qm_dot_pb_dot_frontend__pb2.QmDataRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.QmDataResponse.FromString,
                )
        self.PullSimulatorSamples = channel.unary_stream(
                '/qm.grpc.frontend.Frontend/PullSimulatorSamples',
                request_serializer=qm_dot_pb_dot_job__results__pb2.PullSimulatorSamplesRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_job__results__pb2.SimulatorSamplesResponse.FromString,
                )
        self.OpenQuantumMachine = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/OpenQuantumMachine',
                request_serializer=qm_dot_pb_dot_qm__manager__pb2.OpenQuantumMachineRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_qm__manager__pb2.OpenQuantumMachineResponse.FromString,
                )
        self.CloseQuantumMachine = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/CloseQuantumMachine',
                request_serializer=qm_dot_pb_dot_qm__manager__pb2.CloseQuantumMachineRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_qm__manager__pb2.CloseQuantumMachineResponse.FromString,
                )
        self.GetQuantumMachine = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/GetQuantumMachine',
                request_serializer=qm_dot_pb_dot_qm__manager__pb2.GetQuantumMachineRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_qm__manager__pb2.GetQuantumMachineResponse.FromString,
                )
        self.GetRunningJob = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/GetRunningJob',
                request_serializer=qm_dot_pb_dot_qm__manager__pb2.GetRunningJobRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_qm__manager__pb2.GetRunningJobResponse.FromString,
                )
        self.ListOpenQuantumMachines = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/ListOpenQuantumMachines',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=qm_dot_pb_dot_qm__manager__pb2.ListOpenQuantumMachinesResponse.FromString,
                )
        self.CloseAllQuantumMachines = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/CloseAllQuantumMachines',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=qm_dot_pb_dot_qm__manager__pb2.CloseAllQuantumMachinesResponse.FromString,
                )
        self.GetControllers = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/GetControllers',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=qm_dot_pb_dot_qm__manager__pb2.GetControllersResponse.FromString,
                )
        self.IsJobRunning = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/IsJobRunning',
                request_serializer=qm_dot_pb_dot_frontend__pb2.IsJobRunningRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.IsJobRunningResponse.FromString,
                )
        self.IsJobAcquiringData = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/IsJobAcquiringData',
                request_serializer=qm_dot_pb_dot_frontend__pb2.IsJobAcquiringDataRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.IsJobAcquiringDataResponse.FromString,
                )
        self.Simulate = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/Simulate',
                request_serializer=qm_dot_pb_dot_frontend__pb2.SimulationRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.SimulationResponse.FromString,
                )
        self.ClearAllJobResults = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/ClearAllJobResults',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.PerformHalDebugCommand = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/PerformHalDebugCommand',
                request_serializer=qm_dot_pb_dot_frontend__pb2.PerformHalDebugCommandRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.PerformHalDebugCommandResponse.FromString,
                )
        self.GetSimulatedQuantumState = channel.unary_unary(
                '/qm.grpc.frontend.Frontend/GetSimulatedQuantumState',
                request_serializer=qm_dot_pb_dot_frontend__pb2.GetSimulatedQuantumStateRequest.SerializeToString,
                response_deserializer=qm_dot_pb_dot_frontend__pb2.GetSimulatedQuantumStateResponse.FromString,
                )


class FrontendServicer(object):
    """The greeting service definition.
    """

    def GetVersion(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HealthCheck(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Execute(self, request, context):
        """Compile and Run a QUA program
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddToQueue(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Compile(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddCompiledToQueue(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RemovePendingJobs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetPendingJobs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobExecutionStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PausedStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Resume(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Halt(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResetDataProcessing(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ValidateConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Init(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PerformQmRequest(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RequestData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PullSimulatorSamples(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OpenQuantumMachine(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CloseQuantumMachine(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetQuantumMachine(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetRunningJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListOpenQuantumMachines(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CloseAllQuantumMachines(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetControllers(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IsJobRunning(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IsJobAcquiringData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Simulate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ClearAllJobResults(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PerformHalDebugCommand(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSimulatedQuantumState(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_FrontendServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.GetVersion,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
            ),
            'HealthCheck': grpc.unary_unary_rpc_method_handler(
                    servicer.HealthCheck,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.HealthCheckResponse.SerializeToString,
            ),
            'Execute': grpc.unary_unary_rpc_method_handler(
                    servicer.Execute,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.ExecutionRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.ExecutionResponse.SerializeToString,
            ),
            'AddToQueue': grpc.unary_unary_rpc_method_handler(
                    servicer.AddToQueue,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.AddToQueueRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.AddToQueueResponse.SerializeToString,
            ),
            'Compile': grpc.unary_unary_rpc_method_handler(
                    servicer.Compile,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.CompileRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.CompileResponse.SerializeToString,
            ),
            'AddCompiledToQueue': grpc.unary_unary_rpc_method_handler(
                    servicer.AddCompiledToQueue,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.AddCompiledToQueueRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.AddCompiledToQueueResponse.SerializeToString,
            ),
            'RemovePendingJobs': grpc.unary_unary_rpc_method_handler(
                    servicer.RemovePendingJobs,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.JobQueryParams.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.RemovePendingJobsResponse.SerializeToString,
            ),
            'GetPendingJobs': grpc.unary_unary_rpc_method_handler(
                    servicer.GetPendingJobs,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.JobQueryParams.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.GetPendingJobsResponse.SerializeToString,
            ),
            'GetJobExecutionStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobExecutionStatus,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.GetJobExecutionStatusRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.GetJobExecutionStatusResponse.SerializeToString,
            ),
            'PausedStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.PausedStatus,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.PausedStatusRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.PausedStatusResponse.SerializeToString,
            ),
            'Resume': grpc.unary_unary_rpc_method_handler(
                    servicer.Resume,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.ResumeRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.ResumeResponse.SerializeToString,
            ),
            'Halt': grpc.unary_unary_rpc_method_handler(
                    servicer.Halt,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.HaltRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.HaltResponse.SerializeToString,
            ),
            'ResetDataProcessing': grpc.unary_unary_rpc_method_handler(
                    servicer.ResetDataProcessing,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.ResetDataProcessingRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.ResetDataProcessingResponse.SerializeToString,
            ),
            'ValidateConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.ValidateConfig,
                    request_deserializer=qm_dot_pb_dot_inc__qua__config__pb2.QuaConfig.FromString,
                    response_serializer=qm_dot_pb_dot_compiler__pb2.ValidationResponse.SerializeToString,
            ),
            'Init': grpc.unary_unary_rpc_method_handler(
                    servicer.Init,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'PerformQmRequest': grpc.unary_unary_rpc_method_handler(
                    servicer.PerformQmRequest,
                    request_deserializer=qm_dot_pb_dot_inc__qm__api__pb2.HighQmApiRequest.FromString,
                    response_serializer=qm_dot_pb_dot_inc__qm__api__pb2.HighQmApiResponse.SerializeToString,
            ),
            'RequestData': grpc.unary_unary_rpc_method_handler(
                    servicer.RequestData,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.QmDataRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.QmDataResponse.SerializeToString,
            ),
            'PullSimulatorSamples': grpc.unary_stream_rpc_method_handler(
                    servicer.PullSimulatorSamples,
                    request_deserializer=qm_dot_pb_dot_job__results__pb2.PullSimulatorSamplesRequest.FromString,
                    response_serializer=qm_dot_pb_dot_job__results__pb2.SimulatorSamplesResponse.SerializeToString,
            ),
            'OpenQuantumMachine': grpc.unary_unary_rpc_method_handler(
                    servicer.OpenQuantumMachine,
                    request_deserializer=qm_dot_pb_dot_qm__manager__pb2.OpenQuantumMachineRequest.FromString,
                    response_serializer=qm_dot_pb_dot_qm__manager__pb2.OpenQuantumMachineResponse.SerializeToString,
            ),
            'CloseQuantumMachine': grpc.unary_unary_rpc_method_handler(
                    servicer.CloseQuantumMachine,
                    request_deserializer=qm_dot_pb_dot_qm__manager__pb2.CloseQuantumMachineRequest.FromString,
                    response_serializer=qm_dot_pb_dot_qm__manager__pb2.CloseQuantumMachineResponse.SerializeToString,
            ),
            'GetQuantumMachine': grpc.unary_unary_rpc_method_handler(
                    servicer.GetQuantumMachine,
                    request_deserializer=qm_dot_pb_dot_qm__manager__pb2.GetQuantumMachineRequest.FromString,
                    response_serializer=qm_dot_pb_dot_qm__manager__pb2.GetQuantumMachineResponse.SerializeToString,
            ),
            'GetRunningJob': grpc.unary_unary_rpc_method_handler(
                    servicer.GetRunningJob,
                    request_deserializer=qm_dot_pb_dot_qm__manager__pb2.GetRunningJobRequest.FromString,
                    response_serializer=qm_dot_pb_dot_qm__manager__pb2.GetRunningJobResponse.SerializeToString,
            ),
            'ListOpenQuantumMachines': grpc.unary_unary_rpc_method_handler(
                    servicer.ListOpenQuantumMachines,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=qm_dot_pb_dot_qm__manager__pb2.ListOpenQuantumMachinesResponse.SerializeToString,
            ),
            'CloseAllQuantumMachines': grpc.unary_unary_rpc_method_handler(
                    servicer.CloseAllQuantumMachines,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=qm_dot_pb_dot_qm__manager__pb2.CloseAllQuantumMachinesResponse.SerializeToString,
            ),
            'GetControllers': grpc.unary_unary_rpc_method_handler(
                    servicer.GetControllers,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=qm_dot_pb_dot_qm__manager__pb2.GetControllersResponse.SerializeToString,
            ),
            'IsJobRunning': grpc.unary_unary_rpc_method_handler(
                    servicer.IsJobRunning,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.IsJobRunningRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.IsJobRunningResponse.SerializeToString,
            ),
            'IsJobAcquiringData': grpc.unary_unary_rpc_method_handler(
                    servicer.IsJobAcquiringData,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.IsJobAcquiringDataRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.IsJobAcquiringDataResponse.SerializeToString,
            ),
            'Simulate': grpc.unary_unary_rpc_method_handler(
                    servicer.Simulate,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.SimulationRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.SimulationResponse.SerializeToString,
            ),
            'ClearAllJobResults': grpc.unary_unary_rpc_method_handler(
                    servicer.ClearAllJobResults,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'PerformHalDebugCommand': grpc.unary_unary_rpc_method_handler(
                    servicer.PerformHalDebugCommand,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.PerformHalDebugCommandRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.PerformHalDebugCommandResponse.SerializeToString,
            ),
            'GetSimulatedQuantumState': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSimulatedQuantumState,
                    request_deserializer=qm_dot_pb_dot_frontend__pb2.GetSimulatedQuantumStateRequest.FromString,
                    response_serializer=qm_dot_pb_dot_frontend__pb2.GetSimulatedQuantumStateResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'qm.grpc.frontend.Frontend', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Frontend(object):
    """The greeting service definition.
    """

    @staticmethod
    def GetVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/GetVersion',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def HealthCheck(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/HealthCheck',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.HealthCheckResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Execute(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/Execute',
            qm_dot_pb_dot_frontend__pb2.ExecutionRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.ExecutionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddToQueue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/AddToQueue',
            qm_dot_pb_dot_frontend__pb2.AddToQueueRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.AddToQueueResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Compile(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/Compile',
            qm_dot_pb_dot_frontend__pb2.CompileRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.CompileResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddCompiledToQueue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/AddCompiledToQueue',
            qm_dot_pb_dot_frontend__pb2.AddCompiledToQueueRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.AddCompiledToQueueResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RemovePendingJobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/RemovePendingJobs',
            qm_dot_pb_dot_frontend__pb2.JobQueryParams.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.RemovePendingJobsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetPendingJobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/GetPendingJobs',
            qm_dot_pb_dot_frontend__pb2.JobQueryParams.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.GetPendingJobsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetJobExecutionStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/GetJobExecutionStatus',
            qm_dot_pb_dot_frontend__pb2.GetJobExecutionStatusRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.GetJobExecutionStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PausedStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/PausedStatus',
            qm_dot_pb_dot_frontend__pb2.PausedStatusRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.PausedStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Resume(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/Resume',
            qm_dot_pb_dot_frontend__pb2.ResumeRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.ResumeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Halt(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/Halt',
            qm_dot_pb_dot_frontend__pb2.HaltRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.HaltResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ResetDataProcessing(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/ResetDataProcessing',
            qm_dot_pb_dot_frontend__pb2.ResetDataProcessingRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.ResetDataProcessingResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ValidateConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/ValidateConfig',
            qm_dot_pb_dot_inc__qua__config__pb2.QuaConfig.SerializeToString,
            qm_dot_pb_dot_compiler__pb2.ValidationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Init(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/Init',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PerformQmRequest(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/PerformQmRequest',
            qm_dot_pb_dot_inc__qm__api__pb2.HighQmApiRequest.SerializeToString,
            qm_dot_pb_dot_inc__qm__api__pb2.HighQmApiResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RequestData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/RequestData',
            qm_dot_pb_dot_frontend__pb2.QmDataRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.QmDataResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PullSimulatorSamples(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/qm.grpc.frontend.Frontend/PullSimulatorSamples',
            qm_dot_pb_dot_job__results__pb2.PullSimulatorSamplesRequest.SerializeToString,
            qm_dot_pb_dot_job__results__pb2.SimulatorSamplesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def OpenQuantumMachine(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/OpenQuantumMachine',
            qm_dot_pb_dot_qm__manager__pb2.OpenQuantumMachineRequest.SerializeToString,
            qm_dot_pb_dot_qm__manager__pb2.OpenQuantumMachineResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CloseQuantumMachine(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/CloseQuantumMachine',
            qm_dot_pb_dot_qm__manager__pb2.CloseQuantumMachineRequest.SerializeToString,
            qm_dot_pb_dot_qm__manager__pb2.CloseQuantumMachineResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetQuantumMachine(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/GetQuantumMachine',
            qm_dot_pb_dot_qm__manager__pb2.GetQuantumMachineRequest.SerializeToString,
            qm_dot_pb_dot_qm__manager__pb2.GetQuantumMachineResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetRunningJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/GetRunningJob',
            qm_dot_pb_dot_qm__manager__pb2.GetRunningJobRequest.SerializeToString,
            qm_dot_pb_dot_qm__manager__pb2.GetRunningJobResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListOpenQuantumMachines(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/ListOpenQuantumMachines',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            qm_dot_pb_dot_qm__manager__pb2.ListOpenQuantumMachinesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CloseAllQuantumMachines(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/CloseAllQuantumMachines',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            qm_dot_pb_dot_qm__manager__pb2.CloseAllQuantumMachinesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetControllers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/GetControllers',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            qm_dot_pb_dot_qm__manager__pb2.GetControllersResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def IsJobRunning(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/IsJobRunning',
            qm_dot_pb_dot_frontend__pb2.IsJobRunningRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.IsJobRunningResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def IsJobAcquiringData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/IsJobAcquiringData',
            qm_dot_pb_dot_frontend__pb2.IsJobAcquiringDataRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.IsJobAcquiringDataResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Simulate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/Simulate',
            qm_dot_pb_dot_frontend__pb2.SimulationRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.SimulationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ClearAllJobResults(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/ClearAllJobResults',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PerformHalDebugCommand(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/PerformHalDebugCommand',
            qm_dot_pb_dot_frontend__pb2.PerformHalDebugCommandRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.PerformHalDebugCommandResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetSimulatedQuantumState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qm.grpc.frontend.Frontend/GetSimulatedQuantumState',
            qm_dot_pb_dot_frontend__pb2.GetSimulatedQuantumStateRequest.SerializeToString,
            qm_dot_pb_dot_frontend__pb2.GetSimulatedQuantumStateResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import time
from datetime import datetime
from typing import Optional, Union

from qm.QmJob import QmJob
from qm._errors import JobCancelledError
from qm.pb.frontend_pb2 import GetJobExecutionStatusRequest
from qm.pb.inc_qua_pb2 import QuaProgram
from qm.pb.job_manager_pb2_grpc import JobManagerServiceStub
from qm.utils import insert_input_stream_to_job


class QmPendingJob(object):
    """
    A Class describing a job in the execution queue
    """

    def __init__(
        self,
        job_id,
        machine: "qm.QuantumMachine.QuantumMachine",
        qmm: "qm.QuantumMachinesManager.QuantumMachinesManager",
    ):

        self._id = job_id
        self._machine = machine
        self._qmm = qmm
        self._added_user_id = None
        self._time_added = None
        self._set_metadata()
        self._job_manager_client = JobManagerServiceStub(self._qmm._channel)

    def _set_metadata(self):
        response = self._get_execution_status()
        if response.status.HasField("pending"):
            self._added_user_id = response.status.pending.addedBy
            self._time_added = datetime.fromtimestamp(
                response.status.pending.timeAdded.seconds
            )
        elif response.status.HasField("running"):
            self._added_user_id = response.status.running.addedBy
            self._time_added = datetime.fromtimestamp(
                response.status.running.timeAdded.seconds
            )
        elif response.status.HasField("completed"):
            self._added_user_id = response.status.completed.addedBy
            self._time_added = datetime.fromtimestamp(
                response.status.completed.timeAdded.seconds
            )
        elif response.status.HasField("loading"):
            self._added_user_id = response.status.loading.addedBy
            self._time_added = datetime.fromtimestamp(
                response.status.loading.timeAdded.seconds
            )

    def id(self) -> str:
        """
        :return: The id of the job

        """
        return self._id

    @property
    def user_added(self):
        return self._added_user_id

    @property
    def time_added(self) -> datetime:
        return self._time_added

    def position_in_queue(self) -> Optional[int]:
        """
        Returns the current position of the job in the queue

        :return: The position in the queue

        Example:

            >>> pending_job.position_in_queue()

        """
        response = self._get_execution_status()
        if response.status.HasField("pending"):
            return response.status.pending.positionInQueue
        return None  # Job left queue

    def _get_execution_status(self):
        request = GetJobExecutionStatusRequest()
        request.quantumMachineId = self._machine.id
        request.jobId = self._id
        response = self._qmm._frontend.GetJobExecutionStatus(request)
        return response

    def wait_for_execution(self, timeout: Optional[float] = None) -> QmJob:
        """
        Waits for the job to be executed (start running) or until the timeout has elapsed

        :param timeout: Timeout (in seconds) for this operation

        :return: The running ``QmJob``

        """
        start = time.time()
        end = start + max(0.0, timeout) if timeout is not None else None

        while end is None or time.time() < end:
            response = self._get_execution_status()
            if response.status.HasField("pending") or response.status.HasField(
                "loading"
            ):
                time.sleep(0.02)
            elif response.status.HasField("running") or response.status.HasField(
                "completed"
            ):
                return QmJob(self._qmm, self._id, None)
            else:
                raise JobCancelledError(f"job {self._id} was cancelled")
        raise TimeoutError(
            f"job {self._id} was not started in time. Check the QMApp "
            f"for errors and try reloading the version. "
            f"Please also inform QM about this issue. "
        )

    def cancel(self) -> bool:
        """
        Removes the job from the queue

        :return: true if the operation was successful

        Example:

            >>> pending_job.cancel()

        """
        return self._machine.queue.remove_by_id(self._id) > 0

    def insert_input_stream(
        self,
        input_stream: Union[
            QuaProgram.ArrayVarRefExpression, QuaProgram.AnyScalarExpression
        ],
        data,
    ):
        """
        Insert data to the input stream declared in the QUA program.
        The data is then ready to be read by the program using the advance
        input stream QUA statement.

        Multiple data entries can be inserted before the data is read by the program.

        See `Input streams <https://qm-docs.qualang.io/guides/features#input-streams>`__ for more information.

        :param input_stream: The input stream the data is to be inserted to.
        :param data: The data to be inserted. The data's size must match the size of the input stream.
        """
        insert_input_stream_to_job(self, data, input_stream)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qm/pb/qm_manager.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from qm.pb import inc_qua_config_pb2 as qm_dot_pb_dot_inc__qua__config__pb2
from qm.pb import general_messages_pb2 as qm_dot_pb_dot_general__messages__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='qm/pb/qm_manager.proto',
  package='qm.grpc.qm_manager',
  syntax='proto3',
  serialized_options=b'\n\021qm.grpc.qmManagerB\016QmManagerProtoP\001\242\002\003HLW',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16qm/pb/qm_manager.proto\x12\x12qm.grpc.qm_manager\x1a\x1aqm/pb/inc_qua_config.proto\x1a\x1cqm/pb/general_messages.proto\"\x9c\x01\n\x19OpenQuantumMachineRequest\x12-\n\x06\x63onfig\x18\x01 \x01(\x0b\x32\x1d.qm.grpc.qua_config.QuaConfig\x12\x0f\n\x05never\x18\x02 \x01(\x08H\x00\x12\x10\n\x06\x61lways\x18\x03 \x01(\x08H\x00\x12\x12\n\x08ifNeeded\x18\x04 \x01(\x08H\x00\x42\x19\n\x17oneOfCloseOtherMachines\"\x99\x02\n\x1aOpenQuantumMachineResponse\x12\x11\n\tmachineID\x18\x01 \x01(\t\x12\x0f\n\x07success\x18\x02 \x01(\x08\x12K\n\x16\x63onfigValidationErrors\x18\x03 \x03(\x0b\x32+.qm.grpc.qm_manager.ConfigValidationMessage\x12O\n\x18physicalValidationErrors\x18\x04 \x03(\x0b\x32-.qm.grpc.qm_manager.PhysicalValidationMessage\x12\x39\n\x0eopenQmWarnings\x18\x05 \x03(\x0b\x32!.qm.grpc.qm_manager.OpenQmWarning\"/\n\x1a\x43loseQuantumMachineRequest\x12\x11\n\tmachineID\x18\x01 \x01(\t\"f\n\x1b\x43loseQuantumMachineResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x36\n\x06\x65rrors\x18\x02 \x03(\x0b\x32&.qm.grpc.general_messages.ErrorMessage\"-\n\x18GetQuantumMachineRequest\x12\x11\n\tmachineID\x18\x01 \x01(\t\"\xa6\x01\n\x19GetQuantumMachineResponse\x12\x11\n\tmachineID\x18\x01 \x01(\t\x12-\n\x06\x63onfig\x18\x02 \x01(\x0b\x32\x1d.qm.grpc.qua_config.QuaConfig\x12\x0f\n\x07success\x18\x03 \x01(\x08\x12\x36\n\x06\x65rrors\x18\x04 \x03(\x0b\x32&.qm.grpc.general_messages.ErrorMessage\")\n\x14GetRunningJobRequest\x12\x11\n\tmachineID\x18\x01 \x01(\t\"9\n\x15GetRunningJobResponse\x12\x11\n\tmachineID\x18\x01 \x01(\t\x12\r\n\x05jobId\x18\x02 \x01(\t\"5\n\x1fListOpenQuantumMachinesResponse\x12\x12\n\nmachineIDs\x18\x01 \x03(\t\"j\n\x1f\x43loseAllQuantumMachinesResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x36\n\x06\x65rrors\x18\x02 \x03(\x0b\x32&.qm.grpc.general_messages.ErrorMessage\"M\n\x16GetControllersResponse\x12\x33\n\x0b\x63ontrollers\x18\x01 \x03(\x0b\x32\x1e.qm.grpc.qm_manager.Controller\"\x1a\n\nController\x12\x0c\n\x04name\x18\x01 \x01(\t\"~\n\x17\x43onfigValidationMessage\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\r\n\x05group\x18\x02 \x01(\t\x12\x0c\n\x04path\x18\x03 \x01(\t\x12\x35\n\x05level\x18\x04 \x01(\x0e\x32&.qm.grpc.general_messages.MessageLevel\"\x80\x01\n\x19PhysicalValidationMessage\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\r\n\x05group\x18\x02 \x01(\t\x12\x0c\n\x04path\x18\x03 \x01(\t\x12\x35\n\x05level\x18\x04 \x01(\x0e\x32&.qm.grpc.general_messages.MessageLevel\".\n\rOpenQmWarning\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\tB+\n\x11qm.grpc.qmManagerB\x0eQmManagerProtoP\x01\xa2\x02\x03HLWb\x06proto3'
  ,
  dependencies=[qm_dot_pb_dot_inc__qua__config__pb2.DESCRIPTOR,qm_dot_pb_dot_general__messages__pb2.DESCRIPTOR,])




_OPENQUANTUMMACHINEREQUEST = _descriptor.Descriptor(
  name='OpenQuantumMachineRequest',
  full_name='qm.grpc.qm_manager.OpenQuantumMachineRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='config', full_name='qm.grpc.qm_manager.OpenQuantumMachineRequest.config', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='never', full_name='qm.grpc.qm_manager.OpenQuantumMachineRequest.never', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='always', full_name='qm.grpc.qm_manager.OpenQuantumMachineRequest.always', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ifNeeded', full_name='qm.grpc.qm_manager.OpenQuantumMachineRequest.ifNeeded', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='oneOfCloseOtherMachines', full_name='qm.grpc.qm_manager.OpenQuantumMachineRequest.oneOfCloseOtherMachines',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=105,
  serialized_end=261,
)


_OPENQUANTUMMACHINERESPONSE = _descriptor.Descriptor(
  name='OpenQuantumMachineResponse',
  full_name='qm.grpc.qm_manager.OpenQuantumMachineResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='machineID', full_name='qm.grpc.qm_manager.OpenQuantumMachineResponse.machineID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='success', full_name='qm.grpc.qm_manager.OpenQuantumMachineResponse.success', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='configValidationErrors', full_name='qm.grpc.qm_manager.OpenQuantumMachineResponse.configValidationErrors', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='physicalValidationErrors', full_name='qm.grpc.qm_manager.OpenQuantumMachineResponse.physicalValidationErrors', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='openQmWarnings', full_name='qm.grpc.qm_manager.OpenQuantumMachineResponse.openQmWarnings', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=264,
  serialized_end=545,
)


_CLOSEQUANTUMMACHINEREQUEST = _descriptor.Descriptor(
  name='CloseQuantumMachineRequest',
  full_name='qm.grpc.qm_manager.CloseQuantumMachineRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='machineID', full_name='qm.grpc.qm_manager.CloseQuantumMachineRequest.machineID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=547,
  serialized_end=594,
)


_CLOSEQUANTUMMACHINERESPONSE = _descriptor.Descriptor(
  name='CloseQuantumMachineResponse',
  full_name='qm.grpc.qm_manager.CloseQuantumMachineResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='qm.grpc.qm_manager.CloseQuantumMachineResponse.success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='errors', full_name='qm.grpc.qm_manager.CloseQuantumMachineResponse.errors', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=596,
  serialized_end=698,
)


_GETQUANTUMMACHINEREQUEST = _descriptor.Descriptor(
  name='GetQuantumMachineRequest',
  full_name='qm.grpc.qm_manager.GetQuantumMachineRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='machineID', full_name='qm.grpc.qm_manager.GetQuantumMachineRequest.machineID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=700,
  serialized_end=745,
)


_GETQUANTUMMACHINERESPONSE = _descriptor.Descriptor(
  name='GetQuantumMachineResponse',
  full_name='qm.grpc.qm_manager.GetQuantumMachineResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='machineID', full_name='qm.grpc.qm_manager.GetQuantumMachineResponse.machineID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='config', full_name='qm.grpc.qm_manager.GetQuantumMachineResponse.config', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='success', full_name='qm.grpc.qm_manager.GetQuantumMachineResponse.success', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='errors', full_name='qm.grpc.qm_manager.GetQuantumMachineResponse.errors', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=748,
  serialized_end=914,
)


_GETRUNNINGJOBREQUEST = _descriptor.Descriptor(
  name='GetRunningJobRequest',
  full_name='qm.grpc.qm_manager.GetRunningJobRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='machineID', full_name='qm.grpc.qm_manager.GetRunningJobRequest.machineID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=916,
  serialized_end=957,
)


_GETRUNNINGJOBRESPONSE = _descriptor.Descriptor(
  name='GetRunningJobResponse',
  full_name='qm.grpc.qm_manager.GetRunningJobResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='machineID', full_name='qm.grpc.qm_manager.GetRunningJobResponse.machineID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='jobId', full_name='qm.grpc.qm_manager.GetRunningJobResponse.jobId', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=959,
  serialized_end=1016,
)


_LISTOPENQUANTUMMACHINESRESPONSE = _descriptor.Descriptor(
  name='ListOpenQuantumMachinesResponse',
  full_name='qm.grpc.qm_manager.ListOpenQuantumMachinesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='machineIDs', full_name='qm.grpc.qm_manager.ListOpenQuantumMachinesResponse.machineIDs', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1018,
  serialized_end=1071,
)


_CLOSEALLQUANTUMMACHINESRESPONSE = _descriptor.Descriptor(
  name='CloseAllQuantumMachinesResponse',
  full_name='qm.grpc.qm_manager.CloseAllQuantumMachinesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='qm.grpc.qm_manager.CloseAllQuantumMachinesResponse.success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='errors', full_name='qm.grpc.qm_manager.CloseAllQuantumMachinesResponse.errors', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1073,
  serialized_end=1179,
)


_GETCONTROLLERSRESPONSE = _descriptor.Descriptor(
  name='GetControllersResponse',
  full_name='qm.grpc.qm_manager.GetControllersResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='controllers', full_name='qm.grpc.qm_manager.GetControllersResponse.controllers', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1181,
  serialized_end=1258,
)


_CONTROLLER = _descriptor.Descriptor(
  name='Controller',
  full_name='qm.grpc.qm_manager.Controller',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='qm.grpc.qm_manager.Controller.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1260,
  serialized_end=1286,
)


_CONFIGVALIDATIONMESSAGE = _descriptor.Descriptor(
  name='ConfigValidationMessage',
  full_name='qm.grpc.qm_manager.ConfigValidationMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='qm.grpc.qm_manager.ConfigValidationMessage.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group', full_name='qm.grpc.qm_manager.ConfigValidationMessage.group', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path', full_name='qm.grpc.qm_manager.ConfigValidationMessage.path', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='level', full_name='qm.grpc.qm_manager.ConfigValidationMessage.level', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1288,
  serialized_end=1414,
)


_PHYSICALVALIDATIONMESSAGE = _descriptor.Descriptor(
  name='PhysicalValidationMessage',
  full_name='qm.grpc.qm_manager.PhysicalValidationMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='qm.grpc.qm_manager.PhysicalValidationMessage.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group', full_name='qm.grpc.qm_manager.PhysicalValidationMessage.group', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path', full_name='qm.grpc.qm_manager.PhysicalValidationMessage.path', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='level', full_name='qm.grpc.qm_manager.PhysicalValidationMessage.level', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1417,
  serialized_end=1545,
)


_OPENQMWARNING = _descriptor.Descriptor(
  name='OpenQmWarning',
  full_name='qm.grpc.qm_manager.OpenQmWarning',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='qm.grpc.qm_manager.OpenQmWarning.code', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='qm.grpc.qm_manager.OpenQmWarning.message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1547,
  serialized_end=1593,
)

_OPENQUANTUMMACHINEREQUEST.fields_by_name['config'].message_type = qm_dot_pb_dot_inc__qua__config__pb2._QUACONFIG
_OPENQUANTUMMACHINEREQUEST.oneofs_by_name['oneOfCloseOtherMachines'].fields.append(
  _OPENQUANTUMMACHINEREQUEST.fields_by_name['never'])
_OPENQUANTUMMACHINEREQUEST.fields_by_name['never'].containing_oneof = _OPENQUANTUMMACHINEREQUEST.oneofs_by_name['oneOfCloseOtherMachines']
_OPENQUANTUMMACHINEREQUEST.oneofs_by_name['oneOfCloseOtherMachines'].fields.append(
  _OPENQUANTUMMACHINEREQUEST.fields_by_name['always'])
_OPENQUANTUMMACHINEREQUEST.fields_by_name['always'].containing_oneof = _OPENQUANTUMMACHINEREQUEST.oneofs_by_name['oneOfCloseOtherMachines']
_OPENQUANTUMMACHINEREQUEST.oneofs_by_name['oneOfCloseOtherMachines'].fields.append(
  _OPENQUANTUMMACHINEREQUEST.fields_by_name['ifNeeded'])
_OPENQUANTUMMACHINEREQUEST.fields_by_name['ifNeeded'].containing_oneof = _OPENQUANTUMMACHINEREQUEST.oneofs_by_name['oneOfCloseOtherMachines']
_OPENQUANTUMMACHINERESPONSE.fields_by_name['configValidationErrors'].message_type = _CONFIGVALIDATIONMESSAGE
_OPENQUANTUMMACHINERESPONSE.fields_by_name['physicalValidationErrors'].message_type = _PHYSICALVALIDATIONMESSAGE
_OPENQUANTUMMACHINERESPONSE.fields_by_name['openQmWarnings'].message_type = _OPENQMWARNING
_CLOSEQUANTUMMACHINERESPONSE.fields_by_name['errors'].message_type = qm_dot_pb_dot_general__messages__pb2._ERRORMESSAGE
_GETQUANTUMMACHINERESPONSE.fields_by_name['config'].message_type = qm_dot_pb_dot_inc__qua__config__pb2._QUACONFIG
_GETQUANTUMMACHINERESPONSE.fields_by_name['errors'].message_type = qm_dot_pb_dot_general__messages__pb2._ERRORMESSAGE
_CLOSEALLQUANTUMMACHINESRESPONSE.fields_by_name['errors'].message_type = qm_dot_pb_dot_general__messages__pb2._ERRORMESSAGE
_GETCONTROLLERSRESPONSE.fields_by_name['controllers'].message_type = _CONTROLLER
_CONFIGVALIDATIONMESSAGE.fields_by_name['level'].enum_type = qm_dot_pb_dot_general__messages__pb2._MESSAGELEVEL
_PHYSICALVALIDATIONMESSAGE.fields_by_name['level'].enum_type = qm_dot_pb_dot_general__messages__pb2._MESSAGELEVEL
DESCRIPTOR.message_types_by_name['OpenQuantumMachineRequest'] = _OPENQUANTUMMACHINEREQUEST
DESCRIPTOR.message_types_by_name['OpenQuantumMachineResponse'] = _OPENQUANTUMMACHINERESPONSE
DESCRIPTOR.message_types_by_name['CloseQuantumMachineRequest'] = _CLOSEQUANTUMMACHINEREQUEST
DESCRIPTOR.message_types_by_name['CloseQuantumMachineResponse'] = _CLOSEQUANTUMMACHINERESPONSE
DESCRIPTOR.message_types_by_name['GetQuantumMachineRequest'] = _GETQUANTUMMACHINEREQUEST
DESCRIPTOR.message_types_by_name['GetQuantumMachineResponse'] = _GETQUANTUMMACHINERESPONSE
DESCRIPTOR.message_types_by_name['GetRunningJobRequest'] = _GETRUNNINGJOBREQUEST
DESCRIPTOR.message_types_by_name['GetRunningJobResponse'] = _GETRUNNINGJOBRESPONSE
DESCRIPTOR.message_types_by_name['ListOpenQuantumMachinesResponse'] = _LISTOPENQUANTUMMACHINESRESPONSE
DESCRIPTOR.message_types_by_name['CloseAllQuantumMachinesResponse'] = _CLOSEALLQUANTUMMACHINESRESPONSE
DESCRIPTOR.message_types_by_name['GetControllersResponse'] = _GETCONTROLLERSRESPONSE
DESCRIPTOR.message_types_by_name['Controller'] = _CONTROLLER
DESCRIPTOR.message_types_by_name['ConfigValidationMessage'] = _CONFIGVALIDATIONMESSAGE
DESCRIPTOR.message_types_by_name['PhysicalValidationMessage'] = _PHYSICALVALIDATIONMESSAGE
DESCRIPTOR.message_types_by_name['OpenQmWarning'] = _OPENQMWARNING
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

OpenQuantumMachineRequest = _reflection.GeneratedProtocolMessageType('OpenQuantumMachineRequest', (_message.Message,), {
  'DESCRIPTOR' : _OPENQUANTUMMACHINEREQUEST,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.OpenQuantumMachineRequest)
  })
_sym_db.RegisterMessage(OpenQuantumMachineRequest)

OpenQuantumMachineResponse = _reflection.GeneratedProtocolMessageType('OpenQuantumMachineResponse', (_message.Message,), {
  'DESCRIPTOR' : _OPENQUANTUMMACHINERESPONSE,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.OpenQuantumMachineResponse)
  })
_sym_db.RegisterMessage(OpenQuantumMachineResponse)

CloseQuantumMachineRequest = _reflection.GeneratedProtocolMessageType('CloseQuantumMachineRequest', (_message.Message,), {
  'DESCRIPTOR' : _CLOSEQUANTUMMACHINEREQUEST,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.CloseQuantumMachineRequest)
  })
_sym_db.RegisterMessage(CloseQuantumMachineRequest)

CloseQuantumMachineResponse = _reflection.GeneratedProtocolMessageType('CloseQuantumMachineResponse', (_message.Message,), {
  'DESCRIPTOR' : _CLOSEQUANTUMMACHINERESPONSE,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.CloseQuantumMachineResponse)
  })
_sym_db.RegisterMessage(CloseQuantumMachineResponse)

GetQuantumMachineRequest = _reflection.GeneratedProtocolMessageType('GetQuantumMachineRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETQUANTUMMACHINEREQUEST,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.GetQuantumMachineRequest)
  })
_sym_db.RegisterMessage(GetQuantumMachineRequest)

GetQuantumMachineResponse = _reflection.GeneratedProtocolMessageType('GetQuantumMachineResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETQUANTUMMACHINERESPONSE,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.GetQuantumMachineResponse)
  })
_sym_db.RegisterMessage(GetQuantumMachineResponse)

GetRunningJobRequest = _reflection.GeneratedProtocolMessageType('GetRunningJobRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETRUNNINGJOBREQUEST,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.GetRunningJobRequest)
  })
_sym_db.RegisterMessage(GetRunningJobRequest)

GetRunningJobResponse = _reflection.GeneratedProtocolMessageType('GetRunningJobResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETRUNNINGJOBRESPONSE,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.GetRunningJobResponse)
  })
_sym_db.RegisterMessage(GetRunningJobResponse)

ListOpenQuantumMachinesResponse = _reflection.GeneratedProtocolMessageType('ListOpenQuantumMachinesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTOPENQUANTUMMACHINESRESPONSE,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.ListOpenQuantumMachinesResponse)
  })
_sym_db.RegisterMessage(ListOpenQuantumMachinesResponse)

CloseAllQuantumMachinesResponse = _reflection.GeneratedProtocolMessageType('CloseAllQuantumMachinesResponse', (_message.Message,), {
  'DESCRIPTOR' : _CLOSEALLQUANTUMMACHINESRESPONSE,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.CloseAllQuantumMachinesResponse)
  })
_sym_db.RegisterMessage(CloseAllQuantumMachinesResponse)

GetControllersResponse = _reflection.GeneratedProtocolMessageType('GetControllersResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETCONTROLLERSRESPONSE,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.GetControllersResponse)
  })
_sym_db.RegisterMessage(GetControllersResponse)

Controller = _reflection.GeneratedProtocolMessageType('Controller', (_message.Message,), {
  'DESCRIPTOR' : _CONTROLLER,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.Controller)
  })
_sym_db.RegisterMessage(Controller)

ConfigValidationMessage = _reflection.GeneratedProtocolMessageType('ConfigValidationMessage', (_message.Message,), {
  'DESCRIPTOR' : _CONFIGVALIDATIONMESSAGE,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.ConfigValidationMessage)
  })
_sym_db.RegisterMessage(ConfigValidationMessage)

PhysicalValidationMessage = _reflection.GeneratedProtocolMessageType('PhysicalValidationMessage', (_message.Message,), {
  'DESCRIPTOR' : _PHYSICALVALIDATIONMESSAGE,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.PhysicalValidationMessage)
  })
_sym_db.RegisterMessage(PhysicalValidationMessage)

OpenQmWarning = _reflection.GeneratedProtocolMessageType('OpenQmWarning', (_message.Message,), {
  'DESCRIPTOR' : _OPENQMWARNING,
  '__module__' : 'qm.pb.qm_manager_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.qm_manager.OpenQmWarning)
  })
_sym_db.RegisterMessage(OpenQmWarning)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



from qm.pb.frontend_pb2 import SimulationRequest
from qm.simulate.interface import SimulatorInterface


class LoopbackInterface(SimulatorInterface):
    """
    Creates a loopback interface for use in
    :class:`~qm.simulate.interface.SimulationConfig`.
    A loopback connects the output of the OPX into it's input. This can be defined
    directly using the ports or through the elements.

    :param list connections:

        List of tuples with loopback connections. Each tuple can be:

            1. Physical connection between ports:

                ``(fromController: str, fromPort: int, toController: str, toPort: int)``

            2. Virtual connection between elements:

                ``(fromQE: str, toQE: str, toQEInput: int)``
    :param int latency: The latency between the OPX outputs and its input.
    :param float noisePower: How much noise to add to the input.

    Example::

    >>> job = qmm.simulate(config, prog, SimulationConfig(
    >>>                   duration=20000,
    >>>                   # loopback from output 1 to input 2 of controller 1:
    >>>                   simulation_interface=LoopbackInterface([("con1", 1, "con1", 2)])
    """

    def __init__(self, connections, latency=24, noisePower=0.0):
        if type(latency) is not int or latency < 0:
            raise Exception("latency must be a positive integer")

        self.latency = latency

        if (
            type(noisePower) is not float and type(noisePower) is not int
        ) or noisePower < 0:
            raise Exception("noisePower must be a positive number")

        self.noisePower = noisePower

        if type(connections) is not list:
            raise Exception("connections argument must be of type list")

        self.connections = list()
        for connection in connections:
            if type(connection) is not tuple:
                raise Exception("each connection must be of type tuple")
            if len(connection) == 4:
                if (
                    type(connection[0]) is not str
                    or type(connection[1]) is not int
                    or type(connection[2]) is not str
                    or type(connection[3]) is not int
                ):
                    raise Exception(
                        "connection should be (fromController, fromPort, toController, toPort)"
                    )
                self.connections.append(connection)
            elif len(connection) == 3:
                if (
                    type(connection[0]) is not str
                    or type(connection[1]) is not str
                    or type(connection[2]) is not int
                ):
                    raise Exception("connection should be (fromQE, toQE, toQEInput)")
                self.connections.append(
                    (connection[0], -1, connection[1], connection[2])
                )
            else:
                raise Exception("connection should be tuple of length 3 or 4")

    def update_simulate_request(self, request: SimulationRequest):
        request.simulate.simulationInterface.loopback.SetInParent()
        request.simulate.simulationInterface.loopback.latency = self.latency
        request.simulate.simulationInterface.loopback.noisePower = self.noisePower
        for connection in self.connections:
            con = request.simulate.simulationInterface.loopback.connections.add()
            con.fromController = connection[0]
            con.fromPort = connection[1]
            con.toController = connection[2]
            con.toPort = connection[3]



import datetime
import sys
import traceback
import types
from typing import Optional, Dict, Any

import numpy as np

from qm import _Program, version
from qm.QuaNodeVisitor import QuaNodeVisitor
from qm.QuaSerializingVisitor import QuaSerializingVisitor
from qm._errors import ConfigSerializationException
from qm.pb.inc_qua_config_pb2 import QuaConfig
from qm.pb.inc_qua_pb2 import QuaProgram
from qm.program import load_config
from qm.program.ConfigBuilder import convert_msg_to_config

SERIALIZATION_VALIDATION_ERROR = "SERIALIZATION VALIDATION ERROR"

LOADED_CONFIG_ERROR = "LOADED CONFIG SERIALIZATION ERROR"

CONFIG_ERROR = "CONFIG SERIALIZATION ERROR"

SERIALIZATION_NOT_COMPLETE = "SERIALIZATION WAS NOT COMPLETE"


def generate_qua_script(prog: _Program, config: Optional[dict] = None) -> str:
    if prog.is_in_scope():
        raise RuntimeError("Can not generate script inside the qua program scope")

    proto_config = None
    if config is not None:
        try:
            proto_config = load_config(config)
        except Exception as e:
            raise RuntimeError(
                "Can not generate script - bad config",
            ) from e

    proto_prog = prog.build(QuaConfig())
    return _generate_qua_script_pb(proto_prog, proto_config, config)


def _generate_qua_script_pb(
    proto_prog, proto_config: Optional, original_config: Optional
):
    extra_info = ""
    serialized_program = ""
    pretty_original_config = None
    pretty_proto_config = None

    if original_config is not None:
        try:
            pretty_original_config = _print_config(original_config)
        except Exception as e:
            trace = traceback.format_exception(*sys.exc_info())
            extra_info = extra_info + _error_string(e, trace, CONFIG_ERROR)
            pretty_original_config = original_config

    if proto_config is not None:
        try:
            normalized_config = convert_msg_to_config(proto_config)
            pretty_proto_config = _print_config(normalized_config)
        except Exception as e:
            trace = traceback.format_exception(*sys.exc_info())
            extra_info = extra_info + _error_string(e, trace, LOADED_CONFIG_ERROR)
            pretty_proto_config = normalized_config

    try:
        visitor = QuaSerializingVisitor()
        visitor.visit(proto_prog)
        serialized_program = visitor.out()

        extra_info = extra_info + _validate_program(proto_prog, serialized_program)
    except Exception as e:
        trace = traceback.format_exception(*sys.exc_info())
        extra_info = extra_info + _error_string(
            e, trace, SERIALIZATION_VALIDATION_ERROR
        )

    return f"""
# Single QUA script generated at {datetime.datetime.now()}
# QUA library version: {version.__version__}

{serialized_program}
{extra_info if extra_info else ""}
config = {pretty_original_config}

loaded_config = {pretty_proto_config}

"""


def _validate_program(old_prog, serialized_program: str) -> Optional[str]:
    generated_mod = types.ModuleType("gen")
    exec(serialized_program, generated_mod.__dict__)
    new_prog = generated_mod.prog.build(QuaConfig())

    new_prog_str = _program_string(new_prog)
    old_prog_str = _program_string(old_prog)

    if new_prog_str != old_prog_str:
        return f"""

####     {SERIALIZATION_NOT_COMPLETE}     ####
#
#  Original   {old_prog_str}
#  Serialized {new_prog_str}
#
################################################

        """

    return ""


def _error_string(e: Exception, trace, error_type: str) -> str:
    return f"""

    ####     {error_type}     ####
    #
    #  {str(e)}
    #
    # Trace:
    #   {str(trace)}
    #
    ################################################

            """


def _program_string(prog) -> str:
    """
    Will create a canonized string representation of the program
    """
    copy = QuaProgram()
    copy.CopyFrom(prog)
    strip_location_visitor = _StripLocationVisitor()
    strip_location_visitor.visit(copy)
    string = str(copy)
    string = string.replace("\n", "")
    return string


def _print_config(config_part: Dict[str, Any], indent_level: int = 1) -> str:
    """
    Formats a python dictionary into an executable string representation.
    Unlike pretty print, it better supports nested dictionaries. Also auto converts
    lists into a more compact form.
    Works recursively.
    :param Dict[str, Any] config_part: The dictionary to format
    :param int indent_level: Internally used by the function to indicate the current
    indention
    :returns str: The string representation of the dictionary.
    """
    if indent_level > 100:
        raise ConfigSerializationException(
            "Reached maximum depth of config pretty print"
        )

    config_part_str = ""
    if len(config_part) > 0:
        config_part_str += "{\n"

        for key, value in config_part.items():
            config_part_str += (
                "    " * indent_level
                + f"'{str(key)}': "
                + _value_to_str(indent_level, value)
            )

        if indent_level > 1:
            # add an indentation and go down a line
            config_part_str += "    " * (indent_level - 1) + "},\n"
        else:
            # in root indent level, no need to add a line
            config_part_str += "}"

    else:
        config_part_str = "{},\n"

    return config_part_str


def _value_to_str(indent_level, value):
    # To support numpy types, we convert them to normal python types:
    if type(value).__module__ == np.__name__:
        value = value.tolist()

    is_long_list = isinstance(value, list) and len(value) > 1

    if isinstance(value, dict):
        return _print_config(value, indent_level + 1)
    elif isinstance(value, str):
        return f"'{value}'" + ",\n"
    elif is_long_list and isinstance(value[0], dict):
        temp_str = "[\n"
        for v in value:
            temp_str += "    " * (indent_level + 1) + f"{str(v)},\n"
        temp_str += "    " * indent_level + "],\n"
        return temp_str
    elif is_long_list:
        # replace it with a compact list
        is_single_value = len(set(value)) == 1

        if is_single_value:
            return f"[{value[0]}] * {len(value)}" + ",\n"
        else:
            return f"{_make_compact_string_from_list(value)}" + ",\n"
    else:
        # python basic data types string representation are valid python
        return str(value) + ",\n"


def _make_compact_string_from_list(list_data):
    """
    Turns a multi-value list into the most compact string representation of it,
    replacing identical consecutive values by list multiplication.
    """
    list_string = ""
    # Indices where the value changes
    changes_indices = []
    for i in range(len(list_data)):
        if (i == len(list_data) - 1) or (list_data[i + 1] != list_data[i]):
            changes_indices.append(i)

    prev_index = -1
    prev_number_of_identical = 0
    for curr_index in changes_indices:
        number_of_identical = curr_index - prev_index
        if prev_number_of_identical == 0:
            # First item on the list
            if number_of_identical == 1:
                list_string += f"[{list_data[curr_index]}"
            elif number_of_identical > 1:
                list_string += f"[{list_data[curr_index]}] * {number_of_identical}"
            else:
                raise ConfigSerializationException(
                    "number_of_identical can not be negative"
                )

        else:
            if number_of_identical == 1 and prev_number_of_identical == 1:
                list_string += f", {list_data[curr_index]}"
            elif number_of_identical > 1 and prev_number_of_identical == 1:
                list_string += f"] + [{list_data[curr_index]}] * {number_of_identical}"
            elif number_of_identical == 1 and prev_number_of_identical > 1:
                list_string += f" + [{list_data[curr_index]}"
            elif number_of_identical > 1 and prev_number_of_identical > 1:
                list_string += f" + [{list_data[curr_index]}] * {number_of_identical}"
            else:
                raise ConfigSerializationException(
                    "number_of_identical can not be negative"
                )

        prev_index = curr_index
        prev_number_of_identical = number_of_identical
    if prev_number_of_identical == 1:
        list_string += "]"
    return list_string


class _StripLocationVisitor(QuaNodeVisitor):
    """
    Go over all nodes and if they have a location property, we strip it
    """

    def _default_enter(self, node):
        if hasattr(node, "loc"):
            node.loc = "stripped"
        return True

    @staticmethod
    def strip(node):
        _StripLocationVisitor().visit(node)
